@@ -19,16 +19,8 @@
 # The following S3 methods are registered on load if dplyr is present
 
 collect.arrow_dplyr_query <- function(x, as_data_frame = TRUE, ...) {
-  # head and tail are not ExecNodes, at best we can handle them via sink node
-  # so if there are any steps done after head/tail, we need to
-  # evaluate the query up to then and then do a new query for the rest
-  if (is_collapsed(x) && has_head_tail(x$.data)) {
-    x$.data <- as_adq(dplyr::compute(x$.data))$.data
-  }
-
-  # See query-engine.R for ExecPlan/Nodes
   tryCatch(
-    tab <- do_exec_plan(x),
+    out <- as_arrow_table(x),
     # n = 4 because we want the error to show up as being from collect()
     # and not handle_csv_read_error()
     error = function(e, call = caller_env(n = 4)) {
@@ -37,11 +29,9 @@ collect.arrow_dplyr_query <- function(x, as_data_frame = TRUE, ...) {
   )
 
   if (as_data_frame) {
-    df <- as.data.frame(tab)
-    restore_dplyr_features(df, x)
-  } else {
-    restore_dplyr_features(tab, x)
+    out <- as.data.frame(out)
   }
+  restore_dplyr_features(out, x)
 }
 collect.ArrowTabular <- function(x, as_data_frame = TRUE, ...) {
   if (as_data_frame) {
