@@ -16,8 +16,8 @@
 // under the License.
 
 #include "./arrow_types.h"
+#include "./safe-call-into-r.h"
 
-#if defined(ARROW_R_WITH_ARROW)
 #include <arrow/ipc/reader.h>
 #include <arrow/table.h>
 
@@ -55,6 +55,50 @@ std::shared_ptr<arrow::RecordBatchReader> RecordBatchReader__from_batches(
   }
 }
 
+class RFunctionRecordBatchReader : public arrow::RecordBatchReader {
+ public:
+  RFunctionRecordBatchReader(cpp11::sexp fun,
+                             const std::shared_ptr<arrow::Schema>& schema)
+      : fun_(fun), schema_(schema) {}
+
+  std::shared_ptr<arrow::Schema> schema() const { return schema_; }
+
+  arrow::Status ReadNext(std::shared_ptr<arrow::RecordBatch>* batch_out) {
+    auto batch = SafeCallIntoR<std::shared_ptr<arrow::RecordBatch>>([&]() {
+      cpp11::sexp result_sexp = fun_();
+      if (result_sexp == R_NilValue) {
+        return std::shared_ptr<arrow::RecordBatch>(nullptr);
+      } else if (!Rf_inherits(result_sexp, "RecordBatch")) {
+        cpp11::stop("Expected fun() to return an arrow::RecordBatch");
+      }
+
+      return cpp11::as_cpp<std::shared_ptr<arrow::RecordBatch>>(result_sexp);
+    });
+
+    RETURN_NOT_OK(batch);
+
+    if (batch.ValueUnsafe().get() != nullptr &&
+        !batch.ValueUnsafe()->schema()->Equals(schema_)) {
+      return arrow::Status::Invalid("Expected fun() to return batch with schema '",
+                                    schema_->ToString(), "' but got batch with schema '",
+                                    batch.ValueUnsafe()->schema()->ToString(), "'");
+    }
+
+    *batch_out = batch.ValueUnsafe();
+    return arrow::Status::OK();
+  }
+
+ private:
+  cpp11::function fun_;
+  std::shared_ptr<arrow::Schema> schema_;
+};
+
+// [[arrow::export]]
+std::shared_ptr<arrow::RecordBatchReader> RecordBatchReader__from_function(
+    cpp11::sexp fun_sexp, const std::shared_ptr<arrow::Schema>& schema) {
+  return std::make_shared<RFunctionRecordBatchReader>(fun_sexp, schema);
+}
+
 // [[arrow::export]]
 std::shared_ptr<arrow::RecordBatchReader> RecordBatchReader__from_Table(
     const std::shared_ptr<arrow::Table>& table) {
@@ -68,7 +112,7 @@ std::shared_ptr<arrow::Table> Table__from_RecordBatchReader(
 }
 
 // [[arrow::export]]
-std::shared_ptr<arrow::Table> RecordBatchReader__Head(
+std::shared_ptr<arrow::RecordBatchReader> RecordBatchReader__Head(
     const std::shared_ptr<arrow::RecordBatchReader>& reader, int64_t num_rows) {
   std::vector<std::shared_ptr<arrow::RecordBatch>> batches;
   std::shared_ptr<arrow::RecordBatch> this_batch;
@@ -78,7 +122,8 @@ std::shared_ptr<arrow::Table> RecordBatchReader__Head(
     batches.push_back(this_batch->Slice(0, num_rows));
     num_rows -= this_batch->num_rows();
   }
-  return ValueOrStop(arrow::Table::FromRecordBatches(reader->schema(), batches));
+  return ValueOrStop(
+      arrow::RecordBatchReader::Make(std::move(batches), reader->schema()));
 }
 
 // -------- RecordBatchStreamReader
@@ -148,5 +193,3 @@ cpp11::list ipc___RecordBatchFileReader__batches(
 
   return arrow::r::to_r_list(res);
 }
-
-#endif
