@@ -292,6 +292,25 @@ test_that("more informative error when reading a CSV with headers and schema", {
   )
 })
 
+test_that("read_csv_arrow() and write_csv_arrow() accept connection objects", {
+  skip_if_not(CanRunWithCapturedR())
+
+  tf <- tempfile()
+  on.exit(unlink(tf))
+
+  # make this big enough that we might expose concurrency problems,
+  # but not so big that it slows down the tests
+  test_tbl <- tibble::tibble(
+    x = 1:1e4,
+    y = vapply(x, rlang::hash, character(1), USE.NAMES = FALSE),
+    z = vapply(y, rlang::hash, character(1), USE.NAMES = FALSE)
+  )
+
+  write_csv_arrow(test_tbl, file(tf))
+  expect_identical(read_csv_arrow(tf), test_tbl)
+  expect_identical(read_csv_arrow(file(tf)), read_csv_arrow(tf))
+})
+
 test_that("CSV reader works on files with non-UTF-8 encoding", {
   strings <- c("a", "\u00e9", "\U0001f4a9")
   file_string <- paste0(
@@ -389,7 +408,7 @@ test_that("Write a CSV file with invalid input type", {
   bad_input <- Array$create(1:5)
   expect_error(
     write_csv_arrow(bad_input, csv_file),
-    regexp = "x must be an object of class 'data.frame', 'RecordBatch', or 'Table', not 'Array'."
+    regexp = "x must be an object of class .* not 'Array'."
   )
 })
 
@@ -461,9 +480,12 @@ test_that("Writing a CSV errors when unsupported (yet) readr args are used", {
       append = FALSE,
       quote = "all",
       escape = "double",
-      eol = "\n", ),
-    paste("The following arguments are not yet supported in Arrow: \"append\",",
-          "\"quote\", \"escape\", and \"eol\"")
+      eol = "\n"
+    ),
+    paste(
+      "The following arguments are not yet supported in Arrow: \"append\",",
+      "\"quote\", \"escape\", and \"eol\""
+    )
   )
 })
 
@@ -471,8 +493,10 @@ test_that("write_csv_arrow deals with duplication in sink/file", {
   # errors when both file and sink are supplied
   expect_error(
     write_csv_arrow(tbl, file = csv_file, sink = csv_file),
-    paste("You have supplied both \"file\" and \"sink\" arguments. Please",
-          "supply only one of them")
+    paste(
+      "You have supplied both \"file\" and \"sink\" arguments. Please",
+      "supply only one of them"
+    )
   )
 })
 
@@ -484,8 +508,10 @@ test_that("write_csv_arrow deals with duplication in include_headers/col_names",
       include_header = TRUE,
       col_names = TRUE
     ),
-    paste("You have supplied both \"col_names\" and \"include_header\"",
-          "arguments. Please supply only one of them")
+    paste(
+      "You have supplied both \"col_names\" and \"include_header\"",
+      "arguments. Please supply only one of them"
+    )
   )
 
   written_tbl <- suppressMessages(
@@ -503,3 +529,84 @@ test_that("read_csv_arrow() deals with BOMs (byte-order-marks) correctly", {
     tibble(a = 1, b = 2)
   )
 })
+
+test_that("write_csv_arrow can write from Dataset objects", {
+  skip_if_not_available("dataset")
+  data_dir <- make_temp_dir()
+  write_dataset(tbl_no_dates, data_dir, partitioning = "lgl")
+  data_in <- open_dataset(data_dir)
+
+  csv_file <- tempfile()
+  tbl_out <- write_csv_arrow(data_in, csv_file)
+  expect_true(file.exists(csv_file))
+
+  tbl_in <- read_csv_arrow(csv_file)
+  expect_named(tbl_in, c("dbl", "false", "chr", "lgl"))
+  expect_equal(nrow(tbl_in), 10)
+})
+
+test_that("write_csv_arrow can write from RecordBatchReader objects", {
+  skip_if_not_available("dataset")
+  library(dplyr, warn.conflicts = FALSE)
+
+  query_obj <- arrow_table(tbl_no_dates) %>%
+    filter(lgl == TRUE)
+
+  csv_file <- tempfile()
+  on.exit(unlink(csv_file))
+  tbl_out <- write_csv_arrow(query_obj, csv_file)
+  expect_true(file.exists(csv_file))
+
+  tbl_in <- read_csv_arrow(csv_file)
+  expect_named(tbl_in, c("dbl", "lgl", "false", "chr"))
+  expect_equal(nrow(tbl_in), 3)
+})
+
+test_that("read/write compressed file successfully", {
+  skip_if_not_available("gzip")
+  tfgz <- tempfile(fileext = ".csv.gz")
+  tf <- tempfile(fileext = ".csv")
+
+  write_csv_arrow(tbl, tf)
+  write_csv_arrow(tbl, tfgz)
+  expect_lt(file.size(tfgz), file.size(tf))
+
+  expect_identical(
+    read_csv_arrow(tfgz),
+    tbl
+  )
+  skip_if_not_available("lz4")
+  tflz4 <- tempfile(fileext = ".csv.lz4")
+  write_csv_arrow(tbl, tflz4)
+  expect_false(file.size(tfgz) == file.size(tflz4))
+  expect_identical(
+    read_csv_arrow(tflz4),
+    tbl
+  )
+})
+
+test_that("read/write compressed filesystem path", {
+  skip_if_not_available("zstd")
+  tfzst <- tempfile(fileext = ".csv.zst")
+  fs <- LocalFileSystem$create()$path(tfzst)
+  write_csv_arrow(tbl, fs)
+
+  tf <- tempfile(fileext = ".csv")
+  write_csv_arrow(tbl, tf)
+  expect_lt(file.size(tfzst), file.size(tf))
+  expect_identical(
+    read_csv_arrow(fs),
+    tbl
+  )
+})
+
+test_that("read_csv_arrow() can read sub-second timestamps with col_types T setting (ARROW-15599)", {
+  tbl <- tibble::tibble(time = c("2018-10-07 19:04:05.000", "2018-10-07 19:04:05.001"))
+  tf <- tempfile()
+  on.exit(unlink(tf))
+  write.csv(tbl, tf, row.names = FALSE)
+
+  df <- read_csv_arrow(tf, col_types = "T", col_names = "time", skip = 1)
+  expected <- as.POSIXct(tbl$time, tz = "UTC")
+  expect_equal(df$time, expected, ignore_attr = "tzone")
+})
