@@ -15,7 +15,7 @@
 # specific language governing permissions and limitations
 # under the License.
 
-skip_if_not_available("dataset")
+skip_if(on_old_windows())
 
 library(dplyr, warn.conflicts = FALSE)
 
@@ -23,118 +23,118 @@ library(dplyr, warn.conflicts = FALSE)
 tbl <- slice_sample(example_data_for_sorting, prop = 1L)
 
 test_that("arrange() on integer, double, and character columns", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(int, chr) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(int, desc(dbl)) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(int, desc(desc(dbl))) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(int) %>%
       arrange(desc(dbl)) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(int + dbl, chr) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       mutate(zzz = int + dbl, ) %>%
       arrange(zzz, chr) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       mutate(zzz = int + dbl) %>%
       arrange(int + dbl, chr) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       mutate(int + dbl) %>%
       arrange(int + dbl, chr) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(grp) %>%
       arrange(int, dbl) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(grp) %>%
       arrange(int, dbl, .by_group = TRUE) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(grp, grp2) %>%
       arrange(int, dbl, .by_group = TRUE) %>%
       collect(),
     tbl %>%
       mutate(grp2 = ifelse(is.na(lgl), 1L, as.integer(lgl)))
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(grp) %>%
       arrange(.by_group = TRUE) %>%
       pull(grp),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange() %>%
       collect(),
     tbl %>%
       group_by(grp)
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(grp) %>%
       arrange() %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange() %>%
       collect(),
     tbl
   )
   test_sort_col <- "chr"
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(!!sym(test_sort_col)) %>%
       collect(),
     tbl %>%
       select(chr, lgl)
   )
   test_sort_cols <- c("int", "dbl")
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(!!!syms(test_sort_cols)) %>%
       collect(),
     tbl
@@ -142,14 +142,14 @@ test_that("arrange() on integer, double, and character columns", {
 })
 
 test_that("arrange() on datetime columns", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(dttm, int) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(dttm) %>%
       collect(),
     tbl %>%
@@ -158,8 +158,8 @@ test_that("arrange() on datetime columns", {
 })
 
 test_that("arrange() on logical columns", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       arrange(lgl, int) %>%
       collect(),
     tbl
