@@ -15,7 +15,7 @@
 # specific language governing permissions and limitations
 # under the License.
 
-skip_if_not_available("dataset")
+skip_if(on_old_windows())
 
 library(dplyr, warn.conflicts = FALSE)
 library(stringr)
@@ -23,8 +23,8 @@ library(stringr)
 tbl <- example_data
 
 test_that("group_by groupings are recorded", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(chr) %>%
       select(int, chr) %>%
       filter(int > 5) %>%
@@ -34,20 +34,20 @@ test_that("group_by groupings are recorded", {
 })
 
 test_that("group_by supports creating/renaming", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(chr, numbers = int) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(chr, numbers = int * 4) %>%
       collect(),
     tbl
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(int > 4, lgl, foo = int > 5) %>%
       collect(),
     tbl
@@ -55,8 +55,8 @@ test_that("group_by supports creating/renaming", {
 })
 
 test_that("ungroup", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(chr) %>%
       select(int, chr) %>%
       ungroup() %>%
@@ -66,11 +66,11 @@ test_that("ungroup", {
   )
 
   # to confirm that the above expectation is actually testing what we think it's
-  # testing, verify that expect_dplyr_equal() distinguishes between grouped and
+  # testing, verify that compare_dplyr_binding() distinguishes between grouped and
   # ungrouped tibbles
   expect_error(
-    expect_dplyr_equal(
-      input %>%
+    compare_dplyr_binding(
+      .input %>%
         group_by(chr) %>%
         select(int, chr) %>%
         (function(x) if (inherits(x, "tbl_df")) ungroup(x) else x) %>%
@@ -82,8 +82,8 @@ test_that("ungroup", {
 })
 
 test_that("group_by then rename", {
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(chr) %>%
       select(string = chr, int) %>%
       collect(),
@@ -93,14 +93,14 @@ test_that("group_by then rename", {
 
 test_that("group_by with .drop", {
   test_groups <- c("starting_a_fight", "consoling_a_child", "petting_a_dog")
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(!!!syms(test_groups), .drop = TRUE) %>%
       collect(),
     example_with_logical_factors
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(!!!syms(test_groups), .drop = FALSE) %>%
       collect(),
     example_with_logical_factors
@@ -131,28 +131,40 @@ test_that("group_by with .drop", {
       group_by_drop_default(),
     TRUE
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(.drop = FALSE) %>% # no group by vars
       group_by_drop_default(),
     example_with_logical_factors
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by_drop_default(),
     example_with_logical_factors
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(!!!syms(test_groups)) %>%
       group_by_drop_default(),
     example_with_logical_factors
   )
-  expect_dplyr_equal(
-    input %>%
+  compare_dplyr_binding(
+    .input %>%
       group_by(!!!syms(test_groups), .drop = FALSE) %>%
       ungroup() %>%
       group_by_drop_default(),
     example_with_logical_factors
   )
 })
+
+test_that("group_by() with namespaced functions", {
+  compare_dplyr_binding(
+    .input %>%
+      group_by(int > base::sqrt(25)) %>%
+      summarise(mean(dbl, na.rm = TRUE)) %>%
+      # group order is different from dplyr, hence reordering
+      arrange(`int > base::sqrt(25)`) %>%
+      collect(),
+    tbl
+  )
+})
