@@ -15,11 +15,26 @@
 # specific language governing permissions and limitations
 # under the License.
 
-skip_if_not_installed("duckdb", minimum_version = "0.3.1")
-skip_if_not_installed("dbplyr")
 skip_if_not_available("dataset")
 skip_on_cran()
 
+# this test needs to be the first one since all other test blocks are skipped
+# if duckdb is not installed
+test_that("meaningful error message when duckdb is not installed", {
+  # skipping if duckdb is installed since we're testing the to_duckdb function's
+  # complaint when a user tries to call it, but duckdb isn't available
+  skip_if(requireNamespace("duckdb", quietly = TRUE))
+  ds <- InMemoryDataset$create(example_data)
+  expect_error(
+    to_duckdb(ds),
+    regexp = "Please install the `duckdb` package to pass data with `to_duckdb()`.",
+    fixed = TRUE
+  )
+})
+
+skip_if_not_installed("duckdb", minimum_version = "0.3.1")
+skip_if_not_installed("dbplyr")
+
 library(duckdb, quietly = TRUE)
 library(dplyr, warn.conflicts = FALSE)
 
@@ -31,8 +46,11 @@ test_that("to_duckdb", {
       to_duckdb() %>%
       collect() %>%
       # factors don't roundtrip https://github.com/duckdb/duckdb/issues/1879
-      select(!fct),
-    select(example_data, !fct)
+      select(!fct) %>%
+      arrange(int),
+    example_data %>%
+      select(!fct) %>%
+      arrange(int)
   )
 
   expect_identical(
@@ -41,7 +59,8 @@ test_that("to_duckdb", {
       to_duckdb() %>%
       group_by(lgl) %>%
       summarise(mean_int = mean(int, na.rm = TRUE), mean_dbl = mean(dbl, na.rm = TRUE)) %>%
-      collect(),
+      collect() %>%
+      arrange(mean_int),
     tibble::tibble(
       lgl = c(TRUE, NA, FALSE),
       mean_int = c(3, 6.25, 8.5),
@@ -56,7 +75,8 @@ test_that("to_duckdb", {
       group_by(lgl) %>%
       to_duckdb() %>%
       summarise(mean_int = mean(int, na.rm = TRUE), mean_dbl = mean(dbl, na.rm = TRUE)) %>%
-      collect(),
+      collect() %>%
+      arrange(mean_int),
     tibble::tibble(
       lgl = c(TRUE, NA, FALSE),
       mean_int = c(3, 6.25, 8.5),
@@ -66,7 +86,6 @@ test_that("to_duckdb", {
 })
 
 test_that("to_duckdb then to_arrow", {
-  skip("Flaky, unskip when ARROW-14745 is merged")
   ds <- InMemoryDataset$create(example_data)
 
   ds_rt <- ds %>%
@@ -91,11 +110,14 @@ test_that("to_duckdb then to_arrow", {
     filter(int > 5)
 
   expect_identical(
-    collect(ds_rt),
+    ds_rt %>%
+      collect() %>%
+      arrange(int),
     ds %>%
       select(-fct) %>%
       filter(int > 5) %>%
-      collect()
+      collect() %>%
+      arrange(int)
   )
 
   # Now check errors
@@ -113,6 +135,64 @@ test_that("to_duckdb then to_arrow", {
   )
 })
 
+test_that("to_arrow roundtrip, with dataset", {
+  # these will continue to error until 0.3.2 is released
+  # https://github.com/duckdb/duckdb/pull/2957
+  skip_if_not_installed("duckdb", minimum_version = "0.3.2")
+  # With a multi-part dataset
+  tf <- tempfile()
+  new_ds <- rbind(
+    cbind(example_data, part = 1),
+    cbind(example_data, part = 2),
+    cbind(mutate(example_data, dbl = dbl * 3, dbl2 = dbl2 * 3), part = 3),
+    cbind(mutate(example_data, dbl = dbl * 4, dbl2 = dbl2 * 4), part = 4)
+  )
+  write_dataset(new_ds, tf, partitioning = "part")
+
+  ds <- open_dataset(tf)
+
+  expect_identical(
+    ds %>%
+      to_duckdb() %>%
+      select(-fct) %>%
+      mutate(dbl_plus = dbl + 1) %>%
+      to_arrow() %>%
+      filter(int > 5 & part > 1) %>%
+      collect() %>%
+      arrange(part, int) %>%
+      as.data.frame(),
+    ds %>%
+      select(-fct) %>%
+      filter(int > 5 & part > 1) %>%
+      mutate(dbl_plus = dbl + 1) %>%
+      collect() %>%
+      arrange(part, int)
+  )
+})
+
+test_that("to_arrow roundtrip, with dataset (without wrapping)", {
+  # these will continue to error until 0.3.2 is released
+  # https://github.com/duckdb/duckdb/pull/2957
+  skip_if_not_installed("duckdb", minimum_version = "0.3.2")
+  # With a multi-part dataset
+  tf <- tempfile()
+  new_ds <- rbind(
+    cbind(example_data, part = 1),
+    cbind(example_data, part = 2),
+    cbind(mutate(example_data, dbl = dbl * 3, dbl2 = dbl2 * 3), part = 3),
+    cbind(mutate(example_data, dbl = dbl * 4, dbl2 = dbl2 * 4), part = 4)
+  )
+  write_dataset(new_ds, tf, partitioning = "part")
+
+  out <- open_dataset(tf) %>%
+    to_duckdb() %>%
+    select(-fct) %>%
+    mutate(dbl_plus = dbl + 1) %>%
+    to_arrow()
+
+  expect_r6_class(out, "RecordBatchReader")
+})
+
 # The next set of tests use an already-extant connection to test features of
 # persistence and querying against the table without using the `tbl` itself, so
 # we need to create a connection separate from the ephemeral one that is made
@@ -199,7 +279,8 @@ test_that("to_duckdb passing a connection", {
   table_four <- ds %>%
     select(int, lgl, dbl) %>%
     to_duckdb(con = con_separate, auto_disconnect = FALSE)
-  table_four_name <- table_four$ops$x
+  # dbplyr 2.2.0 renames this internal attribute to lazy_query
+  table_four_name <- table_four$ops$x %||% table_four$lazy_query$x
 
   result <- DBI::dbGetQuery(
     con_separate,
