@@ -31,46 +31,25 @@ namespace internal {
 namespace {
 
 template <typename Type, typename offset_type = typename Type::offset_type>
-Status ListValueLength(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
-  using ScalarType = typename TypeTraits<Type>::ScalarType;
-  using OffsetScalarType = typename TypeTraits<Type>::OffsetScalarType;
-
-  if (batch[0].kind() == Datum::ARRAY) {
-    typename TypeTraits<Type>::ArrayType list(batch[0].array());
-    ArrayData* out_arr = out->mutable_array();
-    auto out_values = out_arr->GetMutableValues<offset_type>(1);
-    const offset_type* offsets = list.raw_value_offsets();
-    ::arrow::internal::VisitBitBlocksVoid(
-        list.data()->buffers[0], list.offset(), list.length(),
-        [&](int64_t position) {
-          *out_values++ = offsets[position + 1] - offsets[position];
-        },
-        [&]() { *out_values++ = 0; });
-  } else {
-    const auto& arg0 = batch[0].scalar_as<ScalarType>();
-    if (arg0.is_valid) {
-      checked_cast<OffsetScalarType*>(out->scalar().get())->value =
-          static_cast<offset_type>(arg0.value->length());
-    }
+Status ListValueLength(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
+  const ArraySpan& arr = batch[0].array;
+  ArraySpan* out_arr = out->array_span();
+  auto out_values = out_arr->GetValues<offset_type>(1);
+  const offset_type* offsets = arr.GetValues<offset_type>(1);
+  // Offsets are always well-defined and monotonic, even for null values
+  for (int64_t i = 0; i < arr.length; ++i) {
+    *out_values++ = offsets[i + 1] - offsets[i];
   }
-
   return Status::OK();
 }
 
-Status FixedSizeListValueLength(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+Status FixedSizeListValueLength(KernelContext* ctx, const ExecSpan& batch,
+                                ExecResult* out) {
   auto width = checked_cast<const FixedSizeListType&>(*batch[0].type()).list_size();
-  if (batch[0].kind() == Datum::ARRAY) {
-    const auto& arr = *batch[0].array();
-    ArrayData* out_arr = out->mutable_array();
-    auto* out_values = out_arr->GetMutableValues<int32_t>(1);
-    std::fill(out_values, out_values + arr.length, width);
-  } else {
-    const auto& arg0 = batch[0].scalar_as<FixedSizeListScalar>();
-    if (arg0.is_valid) {
-      checked_cast<Int32Scalar*>(out->scalar().get())->value = width;
-    }
-  }
-
+  const ArraySpan& arr = batch[0].array;
+  ArraySpan* out_arr = out->array_span();
+  int32_t* out_values = out_arr->GetValues<int32_t>(1);
+  std::fill(out_values, out_values + arr.length, width);
   return Status::OK();
 }
 
@@ -81,36 +60,67 @@ const FunctionDoc list_value_length_doc{
      "Null values emit a null in the output."),
     {"lists"}};
 
-template <typename Type, typename IndexType>
-struct ListElementArray {
-  static Status Exec(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
-    using ListArrayType = typename TypeTraits<Type>::ArrayType;
-    using IndexScalarType = typename TypeTraits<IndexType>::ScalarType;
-    const auto& index_scalar = batch[1].scalar_as<IndexScalarType>();
+template <typename ScalarType, typename T = typename ScalarType::ValueType>
+Status GetListElementIndex(const ExecValue& value, T* out) {
+  if (value.is_scalar()) {
+    const auto& index_scalar = value.scalar_as<ScalarType>();
     if (ARROW_PREDICT_FALSE(!index_scalar.is_valid)) {
       return Status::Invalid("Index must not be null");
     }
-    ListArrayType list_array(batch[0].array());
-    auto index = index_scalar.value;
-    if (ARROW_PREDICT_FALSE(index < 0)) {
-      return Status::Invalid("Index ", index,
-                             " is out of bounds: should be greater than or equal to 0");
+    *out = index_scalar.value;
+  } else {
+    const ArraySpan& index_array = value.array;
+    if (index_array.length > 1) {
+      return Status::NotImplemented(
+          "list_element not yet implemented for arrays "
+          "of list indices");
     }
+    if (index_array.GetNullCount() > 0) {
+      return Status::Invalid("Index must not contain nulls");
+    }
+    *out = index_array.GetValues<T>(1)[0];
+  }
+  if (ARROW_PREDICT_FALSE(*out < 0)) {
+    return Status::Invalid("Index ", *out,
+                           " is out of bounds: should be greater than or equal to 0");
+  }
+  return Status::OK();
+}
+
+template <typename Type, typename IndexType>
+struct ListElement {
+  using ListArrayType = typename TypeTraits<Type>::ArrayType;
+  using IndexScalarType = typename TypeTraits<IndexType>::ScalarType;
+  using IndexValueType = typename IndexScalarType::ValueType;
+  using offset_type = typename Type::offset_type;
+
+  static Status Exec(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
+    const ArraySpan& list = batch[0].array;
+    const ArraySpan& list_values = list.child_data[0];
+    const offset_type* offsets = list.GetValues<offset_type>(1);
+
+    IndexValueType index = 0;
+    RETURN_NOT_OK(GetListElementIndex<IndexScalarType>(batch[1], &index));
+
     std::unique_ptr<ArrayBuilder> builder;
-    RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(), list_array.value_type(), &builder));
-    RETURN_NOT_OK(builder->Reserve(list_array.length()));
-    for (int i = 0; i < list_array.length(); ++i) {
-      if (list_array.IsNull(i)) {
+
+    const Type* list_type = checked_cast<const Type*>(list.type);
+    RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(), list_type->value_type(), &builder));
+    RETURN_NOT_OK(builder->Reserve(list.length));
+    for (int i = 0; i < list.length; ++i) {
+      if (list.IsNull(i)) {
         RETURN_NOT_OK(builder->AppendNull());
         continue;
       }
-      std::shared_ptr<arrow::Array> value_array = list_array.value_slice(i);
-      auto len = value_array->length();
-      if (ARROW_PREDICT_FALSE(index >= static_cast<typename IndexType::c_type>(len))) {
+
+      const offset_type value_offset = offsets[i];
+      const offset_type value_length = offsets[i + 1] - offsets[i];
+      if (ARROW_PREDICT_FALSE(index >=
+                              static_cast<typename IndexType::c_type>(value_length))) {
         return Status::Invalid("Index ", index, " is out of bounds: should be in [0, ",
-                               len, ")");
+                               value_length, ")");
       }
-      RETURN_NOT_OK(builder->AppendArraySlice(*value_array->data(), index, 1));
+      RETURN_NOT_OK(builder->AppendArraySlice(list_values, value_offset + index, 1));
     }
     ARROW_ASSIGN_OR_RAISE(auto result, builder->Finish());
     out->value = result->data();
@@ -118,41 +128,51 @@ struct ListElementArray {
   }
 };
 
-template <typename, typename IndexType>
-struct ListElementScalar {
-  static Status Exec(KernelContext* /*ctx*/, const ExecBatch& batch, Datum* out) {
-    using IndexScalarType = typename TypeTraits<IndexType>::ScalarType;
-    const auto& index_scalar = batch[1].scalar_as<IndexScalarType>();
-    if (ARROW_PREDICT_FALSE(!index_scalar.is_valid)) {
-      return Status::Invalid("Index must not be null");
+template <typename Type, typename IndexType>
+struct FixedSizeListElement {
+  using IndexScalarType = typename TypeTraits<IndexType>::ScalarType;
+  using IndexValueType = typename IndexScalarType::ValueType;
+
+  static Status Exec(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
+    auto item_size = checked_cast<const FixedSizeListType&>(*batch[0].type()).list_size();
+    const ArraySpan& list = batch[0].array;
+    const ArraySpan& list_values = list.child_data[0];
+
+    IndexValueType index = 0;
+    RETURN_NOT_OK(GetListElementIndex<IndexScalarType>(batch[1], &index));
+
+    std::unique_ptr<ArrayBuilder> builder;
+
+    const Type* list_type = checked_cast<const Type*>(list.type);
+    RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(), list_type->value_type(), &builder));
+    RETURN_NOT_OK(builder->Reserve(list.length));
+    for (int i = 0; i < list.length; ++i) {
+      if (list.IsNull(i)) {
+        RETURN_NOT_OK(builder->AppendNull());
+        continue;
+      }
+      if (ARROW_PREDICT_FALSE(index >=
+                              static_cast<typename IndexType::c_type>(item_size))) {
+        return Status::Invalid("Index ", index, " is out of bounds: should be in [0, ",
+                               item_size, ")");
+      }
+      RETURN_NOT_OK(builder->AppendArraySlice(list_values,
+                                              (list.offset + i) * item_size + index, 1));
     }
-    const auto& list_scalar = batch[0].scalar_as<BaseListScalar>();
-    if (ARROW_PREDICT_FALSE(!list_scalar.is_valid)) {
-      out->value = MakeNullScalar(
-          checked_cast<const BaseListType&>(*batch[0].type()).value_type());
-      return Status::OK();
-    }
-    auto list = list_scalar.value;
-    auto index = index_scalar.value;
-    auto len = list->length();
-    if (ARROW_PREDICT_FALSE(index < 0 ||
-                            index >= static_cast<typename IndexType::c_type>(len))) {
-      return Status::Invalid("Index ", index, " is out of bounds: should be in [0, ", len,
-                             ")");
-    }
-    ARROW_ASSIGN_OR_RAISE(out->value, list->GetScalar(index));
+    ARROW_ASSIGN_OR_RAISE(auto result, builder->Finish());
+    out->value = result->data();
     return Status::OK();
   }
 };
 
-template <typename InListType>
-void AddListElementArrayKernels(ScalarFunction* func) {
+template <typename InListType, template <typename...> class Functor>
+void AddListElementKernels(ScalarFunction* func) {
   for (const auto& index_type : IntTypes()) {
-    auto inputs = {InputType::Array(InListType::type_id), InputType::Scalar(index_type)};
+    auto inputs = {InputType(InListType::type_id), InputType(index_type)};
     auto output = OutputType{ListValuesType};
     auto sig = KernelSignature::Make(std::move(inputs), std::move(output),
                                      /*is_varargs=*/false);
-    auto scalar_exec = GenerateInteger<ListElementArray, InListType>({index_type->id()});
+    auto scalar_exec = GenerateInteger<Functor, InListType>({index_type->id()});
     ScalarKernel kernel{std::move(sig), std::move(scalar_exec)};
     kernel.null_handling = NullHandling::COMPUTED_NO_PREALLOCATE;
     kernel.mem_allocation = MemAllocation::NO_PREALLOCATE;
@@ -160,26 +180,10 @@ void AddListElementArrayKernels(ScalarFunction* func) {
   }
 }
 
-void AddListElementArrayKernels(ScalarFunction* func) {
-  AddListElementArrayKernels<ListType>(func);
-  AddListElementArrayKernels<LargeListType>(func);
-  AddListElementArrayKernels<FixedSizeListType>(func);
-}
-
-void AddListElementScalarKernels(ScalarFunction* func) {
-  for (const auto list_type_id : {Type::LIST, Type::LARGE_LIST, Type::FIXED_SIZE_LIST}) {
-    for (const auto& index_type : IntTypes()) {
-      auto inputs = {InputType::Scalar(list_type_id), InputType::Scalar(index_type)};
-      auto output = OutputType{ListValuesType};
-      auto sig = KernelSignature::Make(std::move(inputs), std::move(output),
-                                       /*is_varargs=*/false);
-      auto scalar_exec = GenerateInteger<ListElementScalar, void>({index_type->id()});
-      ScalarKernel kernel{std::move(sig), std::move(scalar_exec)};
-      kernel.null_handling = NullHandling::COMPUTED_NO_PREALLOCATE;
-      kernel.mem_allocation = MemAllocation::NO_PREALLOCATE;
-      DCHECK_OK(func->AddKernel(std::move(kernel)));
-    }
-  }
+void AddListElementKernels(ScalarFunction* func) {
+  AddListElementKernels<ListType, ListElement>(func);
+  AddListElementKernels<LargeListType, ListElement>(func);
+  AddListElementKernels<FixedSizeListType, FixedSizeListElement>(func);
 }
 
 const FunctionDoc list_element_doc(
@@ -190,9 +194,10 @@ const FunctionDoc list_element_doc(
     {"lists", "index"});
 
 struct StructFieldFunctor {
-  static Status ExecArray(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+  static Status Exec(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
     const auto& options = OptionsWrapper<StructFieldOptions>::Get(ctx);
-    std::shared_ptr<Array> current = batch[0].make_array();
+
+    std::shared_ptr<Array> current = MakeArray(batch[0].array.ToArrayData());
     for (const auto& index : options.indices) {
       RETURN_NOT_OK(CheckIndex(index, *current->type()));
       switch (current->type()->id()) {
@@ -243,43 +248,7 @@ struct StructFieldFunctor {
                                    *current->type());
       }
     }
-    *out = current;
-    return Status::OK();
-  }
-
-  static Status ExecScalar(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
-    const auto& options = OptionsWrapper<StructFieldOptions>::Get(ctx);
-    const std::shared_ptr<Scalar>* current = &batch[0].scalar();
-    for (const auto& index : options.indices) {
-      RETURN_NOT_OK(CheckIndex(index, *(*current)->type));
-      if (!(*current)->is_valid) {
-        // out should already be a null scalar of the appropriate type
-        return Status::OK();
-      }
-
-      switch ((*current)->type->id()) {
-        case Type::STRUCT: {
-          current = &checked_cast<const StructScalar&>(**current).value[index];
-          break;
-        }
-        case Type::DENSE_UNION:
-        case Type::SPARSE_UNION: {
-          const auto& union_scalar = checked_cast<const UnionScalar&>(**current);
-          const auto& union_ty = checked_cast<const UnionType&>(*(*current)->type);
-          if (union_scalar.type_code != union_ty.type_codes()[index]) {
-            // out should already be a null scalar of the appropriate type
-            return Status::OK();
-          }
-          current = &union_scalar.value;
-          break;
-        }
-        default:
-          // Should have been checked in ResolveStructFieldType
-          return Status::TypeError("struct_field: cannot reference child field of type ",
-                                   *(*current)->type);
-      }
-    }
-    *out = *current;
+    out->value = std::move(current->data());
     return Status::OK();
   }
 
@@ -300,28 +269,24 @@ struct StructFieldFunctor {
   }
 };
 
-Result<ValueDescr> ResolveStructFieldType(KernelContext* ctx,
-                                          const std::vector<ValueDescr>& descrs) {
+Result<TypeHolder> ResolveStructFieldType(KernelContext* ctx,
+                                          const std::vector<TypeHolder>& types) {
   const auto& options = OptionsWrapper<StructFieldOptions>::Get(ctx);
-  const std::shared_ptr<DataType>* type = &descrs.front().type;
+  const DataType* type = types.front().type;
   for (const auto& index : options.indices) {
-    RETURN_NOT_OK(StructFieldFunctor::CheckIndex(index, **type));
-    type = &(*type)->field(index)->type();
+    RETURN_NOT_OK(StructFieldFunctor::CheckIndex(index, *type));
+    type = type->field(index)->type().get();
   }
-  return ValueDescr(*type, descrs.front().shape);
+  return type;
 }
 
 void AddStructFieldKernels(ScalarFunction* func) {
-  for (const auto shape : {ValueDescr::ARRAY, ValueDescr::SCALAR}) {
-    for (const auto in_type : {Type::STRUCT, Type::DENSE_UNION, Type::SPARSE_UNION}) {
-      ScalarKernel kernel({InputType(in_type, shape)}, OutputType(ResolveStructFieldType),
-                          shape == ValueDescr::ARRAY ? StructFieldFunctor::ExecArray
-                                                     : StructFieldFunctor::ExecScalar,
-                          OptionsWrapper<StructFieldOptions>::Init);
-      kernel.null_handling = NullHandling::COMPUTED_NO_PREALLOCATE;
-      kernel.mem_allocation = MemAllocation::NO_PREALLOCATE;
-      DCHECK_OK(func->AddKernel(std::move(kernel)));
-    }
+  for (const auto in_type : {Type::STRUCT, Type::DENSE_UNION, Type::SPARSE_UNION}) {
+    ScalarKernel kernel({in_type}, ResolveStructFieldType, StructFieldFunctor::Exec,
+                        OptionsWrapper<StructFieldOptions>::Init);
+    kernel.null_handling = NullHandling::COMPUTED_NO_PREALLOCATE;
+    kernel.mem_allocation = MemAllocation::NO_PREALLOCATE;
+    DCHECK_OK(func->AddKernel(std::move(kernel)));
   }
 }
 
@@ -338,59 +303,45 @@ const FunctionDoc struct_field_doc(
      "An empty list of indices returns the argument unchanged."),
     {"values"}, "StructFieldOptions", /*options_required=*/true);
 
-Result<ValueDescr> MakeStructResolve(KernelContext* ctx,
-                                     const std::vector<ValueDescr>& descrs) {
+Result<TypeHolder> MakeStructResolve(KernelContext* ctx,
+                                     const std::vector<TypeHolder>& types) {
   auto names = OptionsWrapper<MakeStructOptions>::Get(ctx).field_names;
   auto nullable = OptionsWrapper<MakeStructOptions>::Get(ctx).field_nullability;
   auto metadata = OptionsWrapper<MakeStructOptions>::Get(ctx).field_metadata;
 
   if (names.size() == 0) {
-    names.resize(descrs.size());
-    nullable.resize(descrs.size(), true);
-    metadata.resize(descrs.size(), nullptr);
+    names.resize(types.size());
+    nullable.resize(types.size(), true);
+    metadata.resize(types.size(), nullptr);
     int i = 0;
     for (auto& name : names) {
       name = std::to_string(i++);
     }
-  } else if (names.size() != descrs.size() || nullable.size() != descrs.size() ||
-             metadata.size() != descrs.size()) {
-    return Status::Invalid("make_struct() was passed ", descrs.size(), " arguments but ",
+  } else if (names.size() != types.size() || nullable.size() != types.size() ||
+             metadata.size() != types.size()) {
+    return Status::Invalid("make_struct() was passed ", types.size(), " arguments but ",
                            names.size(), " field names, ", nullable.size(),
                            " nullability bits, and ", metadata.size(),
                            " metadata dictionaries.");
   }
 
   size_t i = 0;
-  FieldVector fields(descrs.size());
+  FieldVector fields(types.size());
 
-  ValueDescr::Shape shape = ValueDescr::SCALAR;
-  for (const ValueDescr& descr : descrs) {
-    if (descr.shape != ValueDescr::SCALAR) {
-      shape = ValueDescr::ARRAY;
-    } else {
-      switch (descr.type->id()) {
-        case Type::EXTENSION:
-        case Type::DENSE_UNION:
-        case Type::SPARSE_UNION:
-          return Status::NotImplemented("Broadcasting scalars of type ", *descr.type);
-        default:
-          break;
-      }
-    }
-
-    fields[i] =
-        field(std::move(names[i]), descr.type, nullable[i], std::move(metadata[i]));
+  for (const TypeHolder& type : types) {
+    fields[i] = field(std::move(names[i]), type.GetSharedPtr(), nullable[i],
+                      std::move(metadata[i]));
     ++i;
   }
 
-  return ValueDescr{struct_(std::move(fields)), shape};
+  return TypeHolder(struct_(std::move(fields)));
 }
 
-Status MakeStructExec(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
-  ARROW_ASSIGN_OR_RAISE(auto descr, MakeStructResolve(ctx, batch.GetDescriptors()));
+Status MakeStructExec(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
+  ARROW_ASSIGN_OR_RAISE(TypeHolder type, MakeStructResolve(ctx, batch.GetTypes()));
 
   for (int i = 0; i < batch.num_values(); ++i) {
-    const auto& field = checked_cast<const StructType&>(*descr.type).field(i);
+    const auto& field = checked_cast<const StructType&>(*type.type).field(i);
     if (batch[i].null_count() > 0 && !field->nullable()) {
       return Status::Invalid("Output field ", field, " (#", i,
                              ") does not allow nulls but the corresponding "
@@ -398,29 +349,20 @@ Status MakeStructExec(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
     }
   }
 
-  if (descr.shape == ValueDescr::SCALAR) {
-    ScalarVector scalars(batch.num_values());
-    for (int i = 0; i < batch.num_values(); ++i) {
-      scalars[i] = batch[i].scalar();
-    }
-
-    *out =
-        Datum(std::make_shared<StructScalar>(std::move(scalars), std::move(descr.type)));
-    return Status::OK();
-  }
-
-  ArrayVector arrays(batch.num_values());
+  ArrayData* out_data = out->array_data().get();
+  out_data->length = batch.length;
+  out_data->type = type.GetSharedPtr();
+  out_data->child_data.resize(batch.num_values());
   for (int i = 0; i < batch.num_values(); ++i) {
     if (batch[i].is_array()) {
-      arrays[i] = batch[i].make_array();
-      continue;
+      out_data->child_data[i] = batch[i].array.ToArrayData();
+    } else {
+      ARROW_ASSIGN_OR_RAISE(
+          std::shared_ptr<Array> promoted,
+          MakeArrayFromScalar(*batch[i].scalar, batch.length, ctx->memory_pool()));
+      out_data->child_data[i] = promoted->data();
     }
-
-    ARROW_ASSIGN_OR_RAISE(arrays[i], MakeArrayFromScalar(*batch[i].scalar(), batch.length,
-                                                         ctx->memory_pool()));
   }
-
-  *out = std::make_shared<StructArray>(descr.type, batch.length, std::move(arrays));
   return Status::OK();
 }
 
@@ -431,31 +373,29 @@ const FunctionDoc make_struct_doc{"Wrap Arrays into a StructArray",
                                   "MakeStructOptions"};
 template <typename KeyType>
 struct MapLookupFunctor {
-  static Result<int64_t> GetOneMatchingIndex(const Array& keys,
-                                             const Scalar& query_key_scalar,
-                                             const bool* from_back) {
+  using UnboxedKey = typename UnboxScalar<KeyType>::T;
+  static Result<int64_t> GetOneMatchingIndex(const ArraySpan& keys, UnboxedKey query_key,
+                                             const bool use_last) {
     int64_t match_index = -1;
-    RETURN_NOT_OK(
-        FindMatchingIndices(keys, query_key_scalar, [&](int64_t index) -> Status {
-          match_index = index;
-          if (*from_back) {
-            return Status::OK();
-          } else {
-            return Status::Cancelled("Found key match for FIRST");
-          }
-        }));
-
+    RETURN_NOT_OK(FindMatchingIndices(keys, query_key, [&](int64_t index) -> Status {
+      match_index = index;
+      if (use_last) {
+        return Status::OK();
+      } else {
+        // If use_last is false, then this will abort the loop
+        return Status::Cancelled("Found match, short-circuiting");
+      }
+    }));
     return match_index;
   }
 
   template <typename FoundItem>
-  static Status FindMatchingIndices(const Array& keys, const Scalar& query_key_scalar,
+  static Status FindMatchingIndices(const ArraySpan& keys, UnboxedKey query_key,
                                     FoundItem callback) {
-    const auto query_key = UnboxScalar<KeyType>::Unbox(query_key_scalar);
     int64_t index = 0;
     Status status = VisitArrayValuesInline<KeyType>(
-        *keys.data(),
-        [&](decltype(query_key) key) -> Status {
+        keys,
+        [&](UnboxedKey key) -> Status {
           if (key == query_key) {
             return callback(index++);
           }
@@ -472,133 +412,97 @@ struct MapLookupFunctor {
     return Status::OK();
   }
 
-  static Status ExecMapArray(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+  static Status Exec(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
     const auto& options = OptionsWrapper<MapLookupOptions>::Get(ctx);
-    const auto& query_key = options.query_key;
-    const auto& occurrence = options.occurrence;
-    const MapArray map_array(batch[0].array());
+    const UnboxedKey query_key = UnboxScalar<KeyType>::Unbox(*options.query_key);
+
+    const ArraySpan& map = batch[0].array;
+    const int32_t* offsets = map.GetValues<int32_t>(1);
+
+    // The struct holding the keys and values may have an offset
+    int64_t kv_offset = map.child_data[0].offset;
+
+    // We create a copy of the keys array because we will adjust the
+    // offset and length for the map probes below
+    ArraySpan map_keys = map.child_data[0].child_data[0];
+    const ArraySpan& map_items = map.child_data[0].child_data[1];
+
+    std::shared_ptr<DataType> item_type =
+        checked_cast<const MapType*>(map.type)->item_type();
 
     std::unique_ptr<ArrayBuilder> builder;
-    if (occurrence == MapLookupOptions::Occurrence::ALL) {
-      RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(),
-                                list(map_array.map_type()->item_type()), &builder));
+    if (options.occurrence == MapLookupOptions::Occurrence::ALL) {
+      RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(), list(item_type), &builder));
       auto list_builder = checked_cast<ListBuilder*>(builder.get());
       auto value_builder = list_builder->value_builder();
 
-      for (int64_t map_array_idx = 0; map_array_idx < map_array.length();
-           ++map_array_idx) {
-        if (!map_array.IsValid(map_array_idx)) {
+      for (int64_t map_index = 0; map_index < map.length; ++map_index) {
+        if (!map.IsValid(map_index)) {
           RETURN_NOT_OK(list_builder->AppendNull());
           continue;
         }
 
-        auto map = map_array.value_slice(map_array_idx);
-        auto keys = checked_cast<const StructArray&>(*map).field(0);
-        auto items = checked_cast<const StructArray&>(*map).field(1);
+        const int64_t item_offset = offsets[map_index] + kv_offset;
+        const int32_t item_size = offsets[map_index + 1] - offsets[map_index];
+
+        // Adjust the keys view to just the map slot that we are about to search
+        map_keys.SetSlice(item_offset, item_size);
+
         bool found_at_least_one_key = false;
-        RETURN_NOT_OK(
-            FindMatchingIndices(*keys, *query_key, [&](int64_t index) -> Status {
-              if (!found_at_least_one_key) RETURN_NOT_OK(list_builder->Append(true));
-              found_at_least_one_key = true;
-              RETURN_NOT_OK(value_builder->AppendArraySlice(*items->data(), index, 1));
-              return Status::OK();
-            }));
+        RETURN_NOT_OK(FindMatchingIndices(map_keys, query_key, [&](int64_t key_index) {
+          if (!found_at_least_one_key) {
+            RETURN_NOT_OK(list_builder->Append(true));
+          }
+          found_at_least_one_key = true;
+          return value_builder->AppendArraySlice(map_items, item_offset + key_index, 1);
+        }));
         if (!found_at_least_one_key) {
+          // Key was not found in this map element, so we append a null list
           RETURN_NOT_OK(list_builder->AppendNull());
         }
       }
       ARROW_ASSIGN_OR_RAISE(auto result, list_builder->Finish());
-      out->value = result->data();
+      out->value = std::move(result->data());
     } else { /* occurrence == FIRST || LAST */
-      RETURN_NOT_OK(
-          MakeBuilder(ctx->memory_pool(), map_array.map_type()->item_type(), &builder));
+      RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(), item_type, &builder));
       RETURN_NOT_OK(builder->Reserve(batch.length));
-      for (int64_t map_array_idx = 0; map_array_idx < map_array.length();
-           ++map_array_idx) {
-        if (!map_array.IsValid(map_array_idx)) {
+      for (int64_t map_index = 0; map_index < map.length; ++map_index) {
+        if (!map.IsValid(map_index)) {
           RETURN_NOT_OK(builder->AppendNull());
           continue;
         }
 
-        auto map = map_array.value_slice(map_array_idx);
-        auto keys = checked_cast<const StructArray&>(*map).field(0);
-        auto items = checked_cast<const StructArray&>(*map).field(1);
-        bool from_back = (occurrence == MapLookupOptions::LAST);
-        ARROW_ASSIGN_OR_RAISE(int64_t key_match_idx,
-                              GetOneMatchingIndex(*keys, *query_key, &from_back));
+        const int64_t item_offset = offsets[map_index] + kv_offset;
+        const int32_t item_size = offsets[map_index + 1] - offsets[map_index];
 
-        if (key_match_idx != -1) {
-          RETURN_NOT_OK(builder->AppendArraySlice(*items->data(), key_match_idx, 1));
+        // Adjust the keys view to just the map slot that we are about to search
+        map_keys.SetSlice(item_offset, item_size);
+
+        ARROW_ASSIGN_OR_RAISE(
+            int64_t item_index,
+            GetOneMatchingIndex(map_keys, query_key,
+                                options.occurrence == MapLookupOptions::LAST));
+
+        if (item_index != -1) {
+          RETURN_NOT_OK(
+              builder->AppendArraySlice(map_items, item_offset + item_index, 1));
         } else {
           RETURN_NOT_OK(builder->AppendNull());
         }
       }
       ARROW_ASSIGN_OR_RAISE(auto result, builder->Finish());
-      out->value = result->data();
-    }
-
-    return Status::OK();
-  }
-
-  static Status ExecMapScalar(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
-    const auto& options = OptionsWrapper<MapLookupOptions>::Get(ctx);
-    const auto& query_key = options.query_key;
-    const auto& occurrence = options.occurrence;
-
-    std::shared_ptr<DataType> item_type =
-        checked_cast<const MapType&>(*batch[0].type()).item_type();
-    const auto& map_scalar = batch[0].scalar_as<MapScalar>();
-
-    if (ARROW_PREDICT_FALSE(!map_scalar.is_valid)) {
-      if (options.occurrence == MapLookupOptions::Occurrence::ALL) {
-        out->value = MakeNullScalar(list(item_type));
-      } else {
-        out->value = MakeNullScalar(item_type);
-      }
-      return Status::OK();
-    }
-
-    const auto& struct_array = checked_cast<const StructArray&>(*map_scalar.value);
-    const std::shared_ptr<Array> keys = struct_array.field(0);
-    const std::shared_ptr<Array> items = struct_array.field(1);
-
-    if (occurrence == MapLookupOptions::Occurrence::ALL) {
-      bool found_at_least_one_key = false;
-      std::unique_ptr<ArrayBuilder> builder;
-      RETURN_NOT_OK(MakeBuilder(ctx->memory_pool(), items->type(), &builder));
-
-      RETURN_NOT_OK(FindMatchingIndices(*keys, *query_key, [&](int64_t index) -> Status {
-        found_at_least_one_key = true;
-        RETURN_NOT_OK(builder->AppendArraySlice(*items->data(), index, 1));
-        return Status::OK();
-      }));
-      if (!found_at_least_one_key) {
-        out->value = MakeNullScalar(list(items->type()));
-      } else {
-        ARROW_ASSIGN_OR_RAISE(auto result, builder->Finish());
-        ARROW_ASSIGN_OR_RAISE(out->value, MakeScalar(list(items->type()), result));
-      }
-    } else { /* occurrence == FIRST || LAST */
-      bool from_back = (occurrence == MapLookupOptions::LAST);
-
-      ARROW_ASSIGN_OR_RAISE(int64_t key_match_idx,
-                            GetOneMatchingIndex(*keys, *query_key, &from_back));
-      if (key_match_idx != -1) {
-        ARROW_ASSIGN_OR_RAISE(out->value, items->GetScalar(key_match_idx));
-      } else {
-        out->value = MakeNullScalar(items->type());
-      }
+      out->value = std::move(result->data());
     }
     return Status::OK();
   }
 };
 
-Result<ValueDescr> ResolveMapLookupType(KernelContext* ctx,
-                                        const std::vector<ValueDescr>& descrs) {
+Result<TypeHolder> ResolveMapLookupType(KernelContext* ctx,
+                                        const std::vector<TypeHolder>& types) {
   const auto& options = OptionsWrapper<MapLookupOptions>::Get(ctx);
-  std::shared_ptr<DataType> type = descrs.front().type;
-  std::shared_ptr<DataType> item_type = checked_cast<const MapType&>(*type).item_type();
-  std::shared_ptr<DataType> key_type = checked_cast<const MapType&>(*type).key_type();
+  const auto& type = checked_cast<const MapType&>(*types.front().type);
+  std::shared_ptr<DataType> item_type = type.item_type();
+  std::shared_ptr<DataType> key_type = type.key_type();
 
   if (!options.query_key) {
     return Status::Invalid("map_lookup: query_key can't be empty.");
@@ -612,23 +516,20 @@ Result<ValueDescr> ResolveMapLookupType(KernelContext* ctx,
   }
 
   if (options.occurrence == MapLookupOptions::Occurrence::ALL) {
-    return ValueDescr(list(item_type), descrs.front().shape);
+    return list(item_type);
   } else { /* occurrence == FIRST || LAST */
-    return ValueDescr(item_type, descrs.front().shape);
+    return item_type;
   }
 }
 
 struct ResolveMapLookup {
   KernelContext* ctx;
-  const ExecBatch& batch;
-  Datum* out;
+  const ExecSpan& batch;
+  ExecResult* out;
 
   template <typename KeyType>
   Status Execute() {
-    if (batch[0].kind() == Datum::SCALAR) {
-      return MapLookupFunctor<KeyType>::ExecMapScalar(ctx, batch, out);
-    }
-    return MapLookupFunctor<KeyType>::ExecMapArray(ctx, batch, out);
+    return MapLookupFunctor<KeyType>::Exec(ctx, batch, out);
   }
 
   template <typename KeyType>
@@ -661,7 +562,7 @@ struct ResolveMapLookup {
     return Status::TypeError("Got unsupported type: ", type.ToString());
   }
 
-  static Status Exec(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+  static Status Exec(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
     ResolveMapLookup visitor{ctx, batch, out};
     return VisitTypeInline(*checked_cast<const MapType&>(*batch[0].type()).key_type(),
                            &visitor);
@@ -689,7 +590,7 @@ const FunctionDoc map_lookup_doc{
 
 void RegisterScalarNested(FunctionRegistry* registry) {
   auto list_value_length = std::make_shared<ScalarFunction>(
-      "list_value_length", Arity::Unary(), &list_value_length_doc);
+      "list_value_length", Arity::Unary(), list_value_length_doc);
   DCHECK_OK(list_value_length->AddKernel({InputType(Type::LIST)}, int32(),
                                          ListValueLength<ListType>));
   DCHECK_OK(list_value_length->AddKernel({InputType(Type::FIXED_SIZE_LIST)}, int32(),
@@ -698,25 +599,24 @@ void RegisterScalarNested(FunctionRegistry* registry) {
                                          ListValueLength<LargeListType>));
   DCHECK_OK(registry->AddFunction(std::move(list_value_length)));
 
-  auto list_element = std::make_shared<ScalarFunction>("list_element", Arity::Binary(),
-                                                       &list_element_doc);
-  AddListElementArrayKernels(list_element.get());
-  AddListElementScalarKernels(list_element.get());
+  auto list_element =
+      std::make_shared<ScalarFunction>("list_element", Arity::Binary(), list_element_doc);
+  AddListElementKernels(list_element.get());
   DCHECK_OK(registry->AddFunction(std::move(list_element)));
 
   auto struct_field =
-      std::make_shared<ScalarFunction>("struct_field", Arity::Unary(), &struct_field_doc);
+      std::make_shared<ScalarFunction>("struct_field", Arity::Unary(), struct_field_doc);
   AddStructFieldKernels(struct_field.get());
   DCHECK_OK(registry->AddFunction(std::move(struct_field)));
 
   auto map_lookup =
-      std::make_shared<ScalarFunction>("map_lookup", Arity::Unary(), &map_lookup_doc);
+      std::make_shared<ScalarFunction>("map_lookup", Arity::Unary(), map_lookup_doc);
   AddMapLookupKernels(map_lookup.get());
   DCHECK_OK(registry->AddFunction(std::move(map_lookup)));
 
   static MakeStructOptions kDefaultMakeStructOptions;
   auto make_struct_function = std::make_shared<ScalarFunction>(
-      "make_struct", Arity::VarArgs(), &make_struct_doc, &kDefaultMakeStructOptions);
+      "make_struct", Arity::VarArgs(), make_struct_doc, &kDefaultMakeStructOptions);
 
   ScalarKernel kernel{KernelSignature::Make({InputType{}}, OutputType{MakeStructResolve},
                                             /*is_varargs=*/true),
