@@ -19,11 +19,13 @@
 
 #include <google/cloud/storage/client.h>
 #include <algorithm>
+#include <chrono>
 
 #include "arrow/buffer.h"
 #include "arrow/filesystem/gcsfs_internal.h"
 #include "arrow/filesystem/path_util.h"
 #include "arrow/filesystem/util_internal.h"
+#include "arrow/io/util_internal.h"
 #include "arrow/result.h"
 #include "arrow/util/checked_cast.h"
 #include "arrow/util/thread_pool.h"
@@ -33,13 +35,23 @@
 
 namespace arrow {
 namespace fs {
-struct GcsCredentials {
-  explicit GcsCredentials(std::shared_ptr<google::cloud::Credentials> c)
-      : credentials(std::move(c)) {}
-
+struct GcsCredentialsHolder {
+  // Constructor needed for make_shared
+  explicit GcsCredentialsHolder(std::shared_ptr<google::cloud::Credentials> credentials)
+      : credentials(std::move(credentials)) {}
   std::shared_ptr<google::cloud::Credentials> credentials;
 };
 
+bool GcsCredentials::Equals(const GcsCredentials& other) const {
+  if (holder_->credentials == other.holder_->credentials) {
+    return true;
+  }
+  return anonymous_ == other.anonymous_ && access_token_ == other.access_token_ &&
+         expiration_ == other.expiration_ &&
+         json_credentials_ == other.json_credentials_ &&
+         target_service_account_ == other.target_service_account_;
+}
+
 namespace {
 
 namespace gcs = google::cloud::storage;
@@ -95,12 +107,10 @@ struct GcsPath {
 class GcsInputStream : public arrow::io::InputStream {
  public:
   explicit GcsInputStream(gcs::ObjectReadStream stream, GcsPath path,
-                          gcs::Generation generation, gcs::ReadFromOffset offset,
-                          gcs::Client client)
+                          gcs::Generation generation, gcs::Client client)
       : stream_(std::move(stream)),
         path_(std::move(path)),
         generation_(generation),
-        offset_(offset.value_or(0)),
         client_(std::move(client)) {}
 
   ~GcsInputStream() override = default;
@@ -115,7 +125,7 @@ class GcsInputStream : public arrow::io::InputStream {
 
   Result<int64_t> Tell() const override {
     if (closed()) return Status::Invalid("Cannot use Tell() on a closed stream");
-    return stream_.tellg() + offset_;
+    return stream_.tellg();
   }
 
   // A gcs::ObjectReadStream can be "born closed".  For small objects the stream returns
@@ -156,7 +166,6 @@ class GcsInputStream : public arrow::io::InputStream {
   mutable gcs::ObjectReadStream stream_;
   GcsPath path_;
   gcs::Generation generation_;
-  std::int64_t offset_;
   gcs::Client client_;
   bool closed_ = false;
 };
@@ -164,9 +173,17 @@ class GcsInputStream : public arrow::io::InputStream {
 class GcsOutputStream : public arrow::io::OutputStream {
  public:
   explicit GcsOutputStream(gcs::ObjectWriteStream stream) : stream_(std::move(stream)) {}
-  ~GcsOutputStream() override = default;
+  ~GcsOutputStream() {
+    if (!closed_) {
+      // The common pattern is to close OutputStreams from destructor in arrow.
+      io::internal::CloseFromDestructor(this);
+    }
+  }
 
   Status Close() override {
+    if (closed_) {
+      return Status::OK();
+    }
     stream_.Close();
     closed_ = true;
     return internal::ToArrowStatus(stream_.last_status());
@@ -297,8 +314,15 @@ google::cloud::Options AsGoogleCloudOptions(const GcsOptions& o) {
   if (!o.endpoint_override.empty()) {
     options.set<gcs::RestEndpointOption>(scheme + "://" + o.endpoint_override);
   }
-  if (o.credentials && o.credentials->credentials) {
-    options.set<google::cloud::UnifiedCredentialsOption>(o.credentials->credentials);
+  if (o.credentials.holder() && o.credentials.holder()->credentials) {
+    options.set<google::cloud::UnifiedCredentialsOption>(
+        o.credentials.holder()->credentials);
+  }
+  if (o.retry_limit_seconds.has_value()) {
+    options.set<gcs::RetryPolicyOption>(
+        gcs::LimitedTimeRetryPolicy(
+            std::chrono::milliseconds(static_cast<int>(*o.retry_limit_seconds * 1000)))
+            .clone());
   }
   return options;
 }
@@ -318,19 +342,44 @@ class GcsFileSystem::Impl {
       return GetFileInfoBucket(path, std::move(meta).status());
     }
     auto meta = client_.GetObjectMetadata(path.bucket, path.object);
-    return GetFileInfoObject(path, meta);
+    Result<FileInfo> info = GetFileInfoObject(path, meta);
+    if (!info.ok() || info->type() != FileType::NotFound) {
+      return info;
+    }
+    // Not found case.  It could be this was written to GCS with a different
+    // "Directory" convention, so if there is at least one object that
+    // matches the prefix we assume it is a directory.
+    std::string canonical = internal::EnsureTrailingSlash(path.object);
+    auto list_result = client_.ListObjects(path.bucket, gcs::Prefix(canonical));
+    if (list_result.begin() != list_result.end()) {
+      // If there is at least one result it indicates this is a directory (at
+      // least one object exists that starts with "path/")
+      return FileInfo(path.full_path, FileType::Directory);
+    }
+    // Return the original not-found info if there was no match.
+    return info;
   }
 
   Result<FileInfoVector> GetFileInfo(const FileSelector& select) {
     ARROW_ASSIGN_OR_RAISE(auto p, GcsPath::FromString(select.base_dir));
-    // Adding the trailing '/' avoids problems with files named 'a', 'ab', 'ac'  where GCS
-    // would return all of them if the prefix is 'a'.
+    // Adding the trailing '/' avoids problems with files named 'a', 'ab', 'ac'  where
+    // GCS would return all of them if the prefix is 'a'.
     const auto canonical = internal::EnsureTrailingSlash(p.object);
-    const auto max_depth = internal::Depth(canonical) + select.max_recursion;
+    // Need to add one level when the object is not empty because all
+    // directories have an extra slash.
+    const auto max_depth =
+        internal::Depth(canonical) + select.max_recursion + !p.object.empty();
     auto prefix = p.object.empty() ? gcs::Prefix() : gcs::Prefix(canonical);
     auto delimiter = select.recursive ? gcs::Delimiter() : gcs::Delimiter("/");
+    // Include trailing delimiters ensures that files matching "directory"
+    // conventions are also included in the listing.
+    // Only included for select.recursive false because a delimiter needs
+    // to be specified.
+    auto include_trailing = select.recursive ? gcs::IncludeTrailingDelimiter(false)
+                                             : gcs::IncludeTrailingDelimiter(true);
     FileInfoVector result;
-    for (auto const& o : client_.ListObjects(p.bucket, prefix, delimiter)) {
+    for (auto const& o :
+         client_.ListObjects(p.bucket, prefix, delimiter, include_trailing)) {
       if (!o.ok()) {
         if (select.allow_not_found &&
             o.status().code() == google::cloud::StatusCode::kNotFound) {
@@ -340,11 +389,11 @@ class GcsFileSystem::Impl {
       }
       // Skip the directory itself from the results, and any result that is "too deep"
       // into the recursion.
-      if (o->name() == p.object || internal::Depth(o->name()) > max_depth) {
+      if (o->name() == canonical || internal::Depth(o->name()) > max_depth) {
         continue;
       }
       auto path = internal::ConcatAbstractPath(o->bucket(), o->name());
-      result.push_back(ToFileInfo(path, *o));
+      result.push_back(ToFileInfo(path, *o, /*normalize_directories=*/true));
     }
     // Finding any elements indicates the directory was found.
     if (!result.empty() || select.allow_not_found) {
@@ -365,7 +414,7 @@ class GcsFileSystem::Impl {
   google::cloud::StatusOr<gcs::ObjectMetadata> CreateDirMarker(const std::string& bucket,
                                                                util::string_view name) {
     // Make the name canonical.
-    const auto canonical = internal::RemoveTrailingSlash(name).to_string();
+    const auto canonical = internal::EnsureTrailingSlash(name);
     google::cloud::StatusOr<gcs::ObjectMetadata> object = client_.InsertObject(
         bucket, canonical, std::string(),
         gcs::WithObjectMetadata(
@@ -398,6 +447,13 @@ class GcsFileSystem::Impl {
       if (o) {
         if (IsDirectory(*o)) break;
         return NotDirectoryError(*o);
+      } else {
+        // If we didn't find the raw path, check if there is an entry
+        // ending in a slash.
+        o = client_.GetObjectMetadata(bucket, internal::EnsureTrailingSlash(dir));
+        if (o) {
+          break;
+        }
       }
       missing_parents.push_back(dir);
     }
@@ -430,15 +486,17 @@ class GcsFileSystem::Impl {
 
   Status CreateDir(const GcsPath& p) {
     if (p.object.empty()) {
-      return internal::ToArrowStatus(
-          client_
-              .CreateBucket(p.bucket, gcs::BucketMetadata().set_location(
-                                          options_.default_bucket_location))
-              .status());
+      auto metadata =
+          gcs::BucketMetadata().set_location(options_.default_bucket_location);
+      return internal::ToArrowStatus(client_.CreateBucket(p.bucket, metadata).status());
     }
     auto parent = p.parent();
     if (!parent.object.empty()) {
-      auto o = client_.GetObjectMetadata(p.bucket, parent.object);
+      auto o = client_.GetObjectMetadata(p.bucket,
+                                         internal::EnsureTrailingSlash(parent.object));
+      if (!o.ok()) {
+        return internal::ToArrowStatus(o.status());
+      }
       if (!IsDirectory(*o)) return NotDirectoryError(*o);
     }
     return internal::ToArrowStatus(CreateDirMarker(p.bucket, p.object).status());
@@ -451,7 +509,8 @@ class GcsFileSystem::Impl {
   Status DeleteDir(const GcsPath& p, const io::IOContext& io_context) {
     RETURN_NOT_OK(DeleteDirContents(p, /*missing_dir_ok=*/false, io_context));
     if (!p.object.empty()) {
-      return internal::ToArrowStatus(client_.DeleteObject(p.bucket, p.object));
+      auto canonical = std::string(internal::EnsureTrailingSlash(p.object));
+      return internal::ToArrowStatus(client_.DeleteObject(p.bucket, canonical));
     }
     return internal::ToArrowStatus(client_.DeleteBucket(p.bucket));
   }
@@ -484,7 +543,7 @@ class GcsFileSystem::Impl {
       submitted.push_back(DeferNotOk(io_context.executor()->Submit(async_delete, o)));
     }
 
-    if (!missing_dir_ok && !at_least_one_obj && !dir) {
+    if (!missing_dir_ok && !at_least_one_obj && !dir && !p.object.empty()) {
       // No files were found and no directory marker exists
       return Status::IOError("No such directory: ", p.full_path);
     }
@@ -544,8 +603,7 @@ class GcsFileSystem::Impl {
                                                            gcs::ReadFromOffset offset) {
     auto stream = client_.ReadObject(path.bucket, path.object, generation, offset);
     ARROW_GCS_RETURN_NOT_OK(stream.status());
-    return std::make_shared<GcsInputStream>(std::move(stream), path, gcs::Generation(),
-                                            offset, client_);
+    return std::make_shared<GcsInputStream>(std::move(stream), path, generation, client_);
   }
 
   Result<std::shared_ptr<io::OutputStream>> OpenOutputStream(
@@ -603,10 +661,25 @@ class GcsFileSystem::Impl {
     return internal::ToArrowStatus(meta.status());
   }
 
+  // The normalize_directories parameter is needed because
+  // how a directory is listed.  If a specific path is asked
+  // for with a trailing slash it is expected to have a trailing
+  // slash [1] but for recursive listings it is expected that
+  // directories have their path normalized [2].
+  // [1]
+  // https://github.com/apache/arrow/blob/3eaa7dd0e8b3dabc5438203331f05e3e6c011e37/python/pyarrow/tests/test_fs.py#L688
+  // [2]
+  // https://github.com/apache/arrow/blob/3eaa7dd0e8b3dabc5438203331f05e3e6c011e37/cpp/src/arrow/filesystem/test_util.cc#L767
   static FileInfo ToFileInfo(const std::string& full_path,
-                             const gcs::ObjectMetadata& meta) {
-    if (IsDirectory(meta)) {
-      return FileInfo(full_path, FileType::Directory);
+                             const gcs::ObjectMetadata& meta,
+                             bool normalize_directories = false) {
+    if (IsDirectory(meta) || (!full_path.empty() && full_path.back() == '/')) {
+      if (normalize_directories) {
+        auto normalized = std::string(internal::RemoveTrailingSlash(full_path));
+        return FileInfo(std::move(normalized), FileType::Directory);
+      } else {
+        return FileInfo(full_path, FileType::Directory);
+      }
     }
     auto info = FileInfo(full_path, FileType::File);
     info.set_size(static_cast<int64_t>(meta.size()));
@@ -621,33 +694,43 @@ class GcsFileSystem::Impl {
   gcs::Client client_;
 };
 
+GcsOptions::GcsOptions() {
+  this->credentials.holder_ = std::make_shared<GcsCredentialsHolder>(
+      google::cloud::MakeGoogleDefaultCredentials());
+  this->scheme = "https";
+}
+
 bool GcsOptions::Equals(const GcsOptions& other) const {
-  return credentials == other.credentials &&
+  return credentials.Equals(other.credentials) &&
          endpoint_override == other.endpoint_override && scheme == other.scheme &&
-         default_bucket_location == other.default_bucket_location;
+         default_bucket_location == other.default_bucket_location &&
+         retry_limit_seconds == other.retry_limit_seconds;
 }
 
 GcsOptions GcsOptions::Defaults() {
-  GcsOptions options{};
-  options.credentials =
-      std::make_shared<GcsCredentials>(google::cloud::MakeGoogleDefaultCredentials());
-  options.scheme = "https";
+  GcsOptions options;
   return options;
 }
 
 GcsOptions GcsOptions::Anonymous() {
   GcsOptions options{};
-  options.credentials =
-      std::make_shared<GcsCredentials>(google::cloud::MakeInsecureCredentials());
+  options.credentials.holder_ =
+      std::make_shared<GcsCredentialsHolder>(google::cloud::MakeInsecureCredentials());
+  options.credentials.anonymous_ = true;
   options.scheme = "http";
   return options;
 }
 
 GcsOptions GcsOptions::FromAccessToken(const std::string& access_token,
-                                       std::chrono::system_clock::time_point expiration) {
+                                       TimePoint expiration) {
   GcsOptions options{};
-  options.credentials = std::make_shared<GcsCredentials>(
-      google::cloud::MakeAccessTokenCredentials(access_token, expiration));
+  options.credentials.holder_ =
+      std::make_shared<GcsCredentialsHolder>(google::cloud::MakeAccessTokenCredentials(
+          access_token,
+          std::chrono::time_point_cast<std::chrono::system_clock::time_point::duration>(
+              expiration)));
+  options.credentials.access_token_ = access_token;
+  options.credentials.expiration_ = expiration;
   options.scheme = "https";
   return options;
 }
@@ -655,17 +738,20 @@ GcsOptions GcsOptions::FromAccessToken(const std::string& access_token,
 GcsOptions GcsOptions::FromImpersonatedServiceAccount(
     const GcsCredentials& base_credentials, const std::string& target_service_account) {
   GcsOptions options{};
-  options.credentials = std::make_shared<GcsCredentials>(
+  options.credentials = base_credentials;
+  options.credentials.holder_ = std::make_shared<GcsCredentialsHolder>(
       google::cloud::MakeImpersonateServiceAccountCredentials(
-          base_credentials.credentials, target_service_account));
+          base_credentials.holder_->credentials, target_service_account));
+  options.credentials.target_service_account_ = target_service_account;
   options.scheme = "https";
   return options;
 }
 
 GcsOptions GcsOptions::FromServiceAccountCredentials(const std::string& json_object) {
   GcsOptions options{};
-  options.credentials = std::make_shared<GcsCredentials>(
+  options.credentials.holder_ = std::make_shared<GcsCredentialsHolder>(
       google::cloud::MakeServiceAccountCredentials(json_object));
+  options.credentials.json_credentials_ = json_object;
   options.scheme = "https";
   return options;
 }
@@ -698,11 +784,16 @@ Result<GcsOptions> GcsOptions::FromUri(const arrow::internal::Uri& uri,
     options_map.emplace(kv.first, kv.second);
   }
 
-  if (!uri.password().empty() || !uri.username().empty()) {
-    return Status::Invalid("GCS does not accept username or password.");
+  const std::string& username = uri.username();
+  bool anonymous = username == "anonymous";
+  if (!username.empty() && !anonymous) {
+    return Status::Invalid("GCS URIs do not accept username except \"anonymous\".");
   }
+  if (!uri.password().empty()) {
+    return Status::Invalid("GCS URIs do not accept password.");
+  }
+  auto options = anonymous ? GcsOptions::Anonymous() : GcsOptions::Defaults();
 
-  auto options = GcsOptions::Defaults();
   for (const auto& kv : options_map) {
     if (kv.first == "location") {
       options.default_bucket_location = kv.second;
@@ -710,6 +801,13 @@ Result<GcsOptions> GcsOptions::FromUri(const arrow::internal::Uri& uri,
       options.scheme = kv.second;
     } else if (kv.first == "endpoint_override") {
       options.endpoint_override = kv.second;
+    } else if (kv.first == "retry_limit_seconds") {
+      double parsed_seconds = atof(kv.second.c_str());
+      if (parsed_seconds <= 0.0) {
+        return Status::Invalid("retry_limit_seconds must be a positive integer, got '",
+                               kv.second, "'");
+      }
+      options.retry_limit_seconds = parsed_seconds;
     } else {
       return Status::Invalid("Unexpected query parameter in GCS URI: '", kv.first, "'");
     }
@@ -726,6 +824,7 @@ Result<GcsOptions> GcsOptions::FromUri(const std::string& uri_string,
 }
 
 std::string GcsFileSystem::type_name() const { return "gcs"; }
+const GcsOptions& GcsFileSystem::options() const { return impl_->options(); }
 
 bool GcsFileSystem::Equals(const FileSystem& other) const {
   if (this == &other) {
@@ -788,6 +887,7 @@ Status GcsFileSystem::CopyFile(const std::string& src, const std::string& dest)
 
 Result<std::shared_ptr<io::InputStream>> GcsFileSystem::OpenInputStream(
     const std::string& path) {
+  ARROW_RETURN_NOT_OK(internal::AssertNoTrailingSlash(path));
   ARROW_ASSIGN_OR_RAISE(auto p, GcsPath::FromString(path));
   return impl_->OpenInputStream(p, gcs::Generation(), gcs::ReadFromOffset());
 }
@@ -798,12 +898,14 @@ Result<std::shared_ptr<io::InputStream>> GcsFileSystem::OpenInputStream(
     return Status::IOError("Cannot open directory '", info.path(),
                            "' as an input stream");
   }
+  ARROW_RETURN_NOT_OK(internal::AssertNoTrailingSlash(info.path()));
   ARROW_ASSIGN_OR_RAISE(auto p, GcsPath::FromString(info.path()));
   return impl_->OpenInputStream(p, gcs::Generation(), gcs::ReadFromOffset());
 }
 
 Result<std::shared_ptr<io::RandomAccessFile>> GcsFileSystem::OpenInputFile(
     const std::string& path) {
+  ARROW_RETURN_NOT_OK(internal::AssertNoTrailingSlash(path));
   ARROW_ASSIGN_OR_RAISE(auto p, GcsPath::FromString(path));
   auto metadata = impl_->GetObjectMetadata(p);
   ARROW_GCS_RETURN_NOT_OK(metadata.status());
@@ -825,6 +927,7 @@ Result<std::shared_ptr<io::RandomAccessFile>> GcsFileSystem::OpenInputFile(
     return Status::IOError("Cannot open directory '", info.path(),
                            "' as an input stream");
   }
+  ARROW_RETURN_NOT_OK(internal::AssertNoTrailingSlash(info.path()));
   ARROW_ASSIGN_OR_RAISE(auto p, GcsPath::FromString(info.path()));
   auto metadata = impl_->GetObjectMetadata(p);
   ARROW_GCS_RETURN_NOT_OK(metadata.status());
@@ -842,6 +945,7 @@ Result<std::shared_ptr<io::RandomAccessFile>> GcsFileSystem::OpenInputFile(
 
 Result<std::shared_ptr<io::OutputStream>> GcsFileSystem::OpenOutputStream(
     const std::string& path, const std::shared_ptr<const KeyValueMetadata>& metadata) {
+  ARROW_RETURN_NOT_OK(internal::AssertNoTrailingSlash(path));
   ARROW_ASSIGN_OR_RAISE(auto p, GcsPath::FromString(path));
   return impl_->OpenOutputStream(p, metadata);
 }
